import os
import sys
import ssl
import time
import json
import struct
import socket
import select
import config
import signal
import logging
from logging import critical as log


def logger(name):
    os.close(2)
    os.open(time.strftime('logs/{0}.%y%m%d'.format(name), time.gmtime()),
            os.O_CREAT | os.O_WRONLY | os.O_APPEND, 0o644)


class Socket():
    # openssl req -x509 -nodes -days 365 -newkey rsa:2048 -keyout key -out cert
    def __init__(self, host, port, keyfile=None):
        self.sock = None
        self.send_eof = False
        self.recv_eof = False

        if str == type(host):
            for i in range(10):
                try:
                    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                    s.connect((host, port))
                    s = ssl.wrap_socket(s)
                    self.sock = s
                    break
                except:
                    time.sleep(1)
        else:
            self.sock = ssl.wrap_socket(host, keyfile, port, True)

        assert(self.sock)

    def send(self, buf):
        assert(self.send_eof is False)

        if not buf:
            buf = b''
            self.send_eof = True

        self.sock.sendall(struct.pack('!I', len(buf)))
        self.sock.sendall(buf)

    def recv(self):
        assert(self.recv_eof is False)

        def recv(n):
            buf = list()
            while n:
                buf.append(self.sock.recv(n))
                if not len(buf[-1]):
                    return

                n -= len(buf[-1])

            return b''.join(buf)

        buf = recv(4)
        if buf:
            l = struct.unpack('!I', buf)[0]
            if l:
                return recv(l)

        self.recv_eof = True

    def __iter__(self):
        return self

    def __next__(self):
        buf = self.recv()

        if buf:
            return buf

        raise StopIteration

    next = __next__


def server(port):
    logging.basicConfig(format='%(asctime)s %(process)s: %(message)s')
    signal.signal(signal.SIGCHLD, signal.SIG_IGN)
    logger('microlight')
    os.dup2(2, 1)

    for cron in config.crons():
        pid = os.fork()
        if pid:
            log('forked pid(%d) cron(%s)', pid, cron[0])
            continue

        logger(cron[0])
        module, function = cron[0].split('.')
        return getattr(__import__(module), function)(*cron[1:])

    for filename in os.listdir('logs'):
        try:
            t = time.mktime(time.strptime(filename.split('.')[-1], '%y%m%d'))
            if time.time() > t + 200000:
                os.remove('logs/' + filename)
                log('removed log file(%s)', filename)
        except:
            log('could not remove log file(%s)', filename)

    with open('/proc/self/cmdline') as fd:
        cmdline = fd.read()

    now = time.time()
    for pid in os.listdir('/proc'):
        try:
            pid = int(pid)
            if pid != os.getpid():
                with open('/proc/{0}/cmdline'.format(pid)) as fd:
                    mtime = os.fstat(fd.fileno()).st_mtime
                    if cmdline == fd.read() and now > mtime + 120:
                        os.kill(pid, signal.SIGKILL)
                        log('killed pid(%d)', pid)
        except:
            pass

    sock = None
    for i in range(10):
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            s.bind(('', port))
            s.listen(5)
            sock = s
            log('listening on port(%s)', port)
            break
        except socket.error:
            time.sleep(1)

    if not sock:
        log('could not bind to port(%s)', port)
        return

    while True:
        conn, addr = sock.accept()

        pid = os.fork()
        if 0 == pid:
            sock.close()
            break

        conn.close()
        log('forked pid(%d) addr%s', pid, addr)

    sock = Socket(conn, 'cert', 'key')
    cmd = json.loads(sock.next())
    module, function = cmd[0].split('.')
    log('addr%s function(%s.%s)', addr, module, function)

    logger('{0}.{1}'.format(module, function))
    for buf in getattr(__import__(module), function)(sock, *cmd[1:]):
        sock.send(buf)


def cli():
    host, port = sys.argv[1].split(':')
    sock = Socket(host, int(port))
    sock.send(json.dumps(sys.argv[2:]).encode('utf8'))

    read_fds = [0, sock.sock]
    while True:
        for fd in select.select(read_fds, [], [])[0]:
            if fd == sock.sock:
                try:
                    os.write(1, sock.next())
                except:
                    return
            else:
                line = sys.stdin.readline()
                if not line:
                    read_fds = [sock.sock]
                sock.send(line.encode('utf8'))


if __name__ == '__main__':
    server(int(sys.argv[1])) if ':' not in sys.argv[1] else cli()
